import logging
import os

from peek_plugin_base.PeekVortexUtil import peekStorageName

logger = logging.getLogger(__name__)


class StorageTestMixin:
    def __init__(self):
        self._dbConn = None

    def setUp(self) -> None:
        from peek_storage_service._private.storage import setupDbConn
        from peek_storage_service._private.storage.DeclarativeBase import metadata
        import peek_storage_service
        from peek_plugin_base.storage.DbConnection import DbConnection
        from peek_storage_service._private.service.PeekStorageConfig import (
            PeekStorageConfig,
        )

        from peek_platform import PeekPlatformConfig

        PeekPlatformConfig.componentName = peekStorageName

        config = PeekStorageConfig()

        alembicDir = os.path.join(
            os.path.dirname(peek_storage_service._private.__file__), "alembic"
        )
        self._dbConn = DbConnection(
            dbConnectString=config.dbConnectString,
            metadata=metadata,
            alembicDir=alembicDir,
            dbEngineArgs=config.dbEngineArgs,
            enableCreateAll=False,
            enableForeignKeys=False,
        )

        self._dbConn.migrate()

    def tearDown(self) -> None:
        self._dbConn.closeAllSessions()



"""add run worker plpy

Revision ID: 72518909970f
Revises: bab14faf0cd7
Create Date: 2020-05-20 20:13:19.092876

"""

# revision identifiers, used by Alembic.

revision = "72518909970f"
down_revision = "bab14faf0cd7"
branch_labels = None
depends_on = None

from alembic import op


def upgrade():
    # This is now moved to objects and created every time Peek starts
    pass


def downgrade():
    pass



sql = """
CREATE OR REPLACE FUNCTION peek_storage_service.load_paylaod_tuples(
	sql_qry_ character varying,
	payload_filt_ character varying,
	loader_module_class_method_ character varying,
	python_path character varying,
	fetch_size_ integer DEFAULT 50)
    RETURNS character varying
    LANGUAGE 'plpython3u'

    COST 100
    VOLATILE 
    
AS $BODY$

import json
from base64 import b64decode
sqlQry = sql_qry_
payloadFilt = json.loads(payload_filt_)
loaderModuleClassMethod = loader_module_class_method_
pythonPath = json.loads(python_path)
fetchSize = fetch_size_

# ---------------
# Setup to use the virtual environment
import sys

sys.path.extend(pythonPath)

# ---------------
# Dynamically load the tuple create method

from importlib.util import find_spec, module_from_spec

modName, className, methodName = loaderModuleClassMethod.rsplit('.',2)

if modName in sys.modules:
	tuple_package = sys.modules[modName]

else:
	modSpec = find_spec(modName)
	if not modSpec:
		 raise Exception("Failed to find package %s,"
						   " is the python package installed?" % modName)
	tuple_package = modSpec.loader.load_module()

TupleClass = getattr(tuple_package, className)
tupleLoaderMethod = getattr(TupleClass, methodName)

# ---------------
# Turn a row["val"] into a row.val
class Wrap:
    row = None
    def __getattr__(self, name):
        return self.row[name]

wrap = Wrap()

# ---------------
# Iterate through and load the tuples
results = []

cursor = plpy.cursor(sqlQry)
while True:
    rows = cursor.fetch(max(500,fetchSize))
    if not rows:
        break
    for row in rows:
        wrap.row = row
        results.append(tupleLoaderMethod(wrap))

# ---------------
# Convert it to a payload

from vortex.Payload import Payload
encodedPayload = Payload(filt=payloadFilt, tuples=results) \
                        .toEncodedPayload() \
                        if results else None

return json.dumps({
    'count': len(results),
    'encodedPayload': encodedPayload
})

$BODY$;

ALTER FUNCTION peek_storage_service.load_paylaod_tuples(character varying, character varying, character varying, character varying, integer)
    OWNER TO peek;


"""



"""add pg load_payload_tuples

Revision ID: 34490abd765c
Revises: 2efc91c0f0f6
Create Date: 2020-04-24 19:47:40.784157

"""

# revision identifiers, used by Alembic.
revision = "34490abd765c"
down_revision = "2efc91c0f0f6"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # This is now moved to objects and created every time Peek starts
    pass


def downgrade():
    pass



"""add plpython3u

Revision ID: 2efc91c0f0f6
Revises:
Create Date: 2020-04-24 19:43:44.230799

"""

# revision identifiers, used by Alembic.
revision = "2efc91c0f0f6"
down_revision = None
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    sql = """
        DROP EXTENSION IF EXISTS plpython3u CASCADE;

        CREATE PROCEDURAL LANGUAGE plpython3u
            HANDLER plpython3_call_handler
            INLINE plpython3_inline_handler
            VALIDATOR plpython3_validator;

        ALTER LANGUAGE plpython3u
            OWNER TO peek;
          """
    op.execute(sql)


def downgrade():
    pass



import logging
from time import sleep

from peek_plugin_base.storage.DbConnection import DbConnection
from psycopg2.extensions import (
    ISOLATION_LEVEL_AUTOCOMMIT,
    ISOLATION_LEVEL_DEFAULT,
)

logger = logging.getLogger(__name__)


class StorageInit:
    def __init__(self, dbConnection: DbConnection):
        self._dbConnection = dbConnection

    def runPreMigrate(self):
        self._upgradeTimescaleDbExtension()

    def runPostMigrate(self):
        from .alembic.objects import object_load_paylaod_tuples
        from .alembic.objects import object_run_generic_python
        from .alembic.objects import object_run_worker_task_python

        session = self._dbConnection.ormSessionCreator()

        objects = (
            object_load_paylaod_tuples,
            object_run_generic_python,
            object_run_worker_task_python,
        )

        for obj in objects:
            logger.debug("(Re)creating object %s", obj.__name__)
            session.execute(obj.sql)

        session.commit()
        session.close()

    def _upgradeTimescaleDbExtension(self):
        rawConn = self._dbConnection.dbEngine.raw_connection()
        rawConn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        cursor = rawConn.cursor()
        try:

            logger.debug("Updating timescaledb extension")
            cursor.execute("ALTER EXTENSION timescaledb UPDATE")

            while len(rawConn.notices) < 2:
                if "shared_preload_libraries" in rawConn.notices:
                    raise Exception(
                        "|shared_preload_libraries = 'timescaledb'| is "
                        "missing from postgresql.conf"
                    )
                sleep(2.0)

            else:
                logger.debug(rawConn.notices[1].replace("\n", ", "))

        except Exception as e:
            if "Start a new session" in str(e):
                logger.debug(
                    "Skipping timescale extension update as the extension is"
                    " already loaded"
                )

                return

            if 'extension "timescaledb" does not exist' in str(e):
                logger.debug(
                    "Skipping timescale extension update as the extension"
                    " doesn't exist yet"
                )

                return

            logger.error("Updating timescaledb extesion failed")
            logger.exception(e)

        finally:
            rawConn.set_isolation_level(ISOLATION_LEVEL_DEFAULT)
            rawConn.close()
            cursor.close()



"""up1_run_generic_py

Revision ID: 51e48a6594d5
Revises: 72518909970f
Create Date: 2020-06-02 22:27:48.650874

"""

# revision identifiers, used by Alembic.
revision = "51e48a6594d5"
down_revision = "72518909970f"
branch_labels = None
depends_on = None

from alembic import op


def upgrade():
    # This is now moved to objects and created every time Peek starts
    pass


def downgrade():
    pass



"""add run_generic_python

Revision ID: 12a0ab3826f3
Revises: 34490abd765c
Create Date: 2020-04-25 15:28:50.279318

"""

# revision identifiers, used by Alembic.
revision = "12a0ab3826f3"
down_revision = "34490abd765c"
branch_labels = None
depends_on = None

from alembic import op


def upgrade():
    # This is now moved to objects and created every time Peek starts
    pass


def downgrade():
    pass



"""
 *  Copyright Synerty HQ Pty Ltd 2013
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by
 *  Synerty HQ Pty Ltd
 *
"""

import logging
from typing import Dict

from sqlalchemy import MetaData

from peek_plugin_base.storage.DbConnection import DbConnection

logger = logging.getLogger(__name__)

dbConn = None


def setupDbConn(
    dbConnectString: str,
    metadata: MetaData,
    alembicDir: str,
    dbEngineArgs: Dict,
):
    global dbConn
    dbConn = DbConnection(
        dbConnectString=dbConnectString,
        metadata=metadata,
        alembicDir=alembicDir,
        dbEngineArgs=dbEngineArgs,
        enableCreateAll=False,
        enableForeignKeys=False,
    )


from . import Setting
